/* (c) Copyright by Tels 1997. All Rights Reserved.
   And I mean ALL RIGHTS!

  [10/16/97] Tels: First version
  [10/20/97] Tels: Rounding problem in filBackpack fixed
             Tels: Counts of 0 are tried, too.
             Tels: speedup: added second method w/o sorting
             Tels: fixed: used was not set to zero while allocating new items
             Tels: fixed: bug in delete all (didnt reset head->next to NULL)
             Tels: speedup: print only better backpacks
  [10/21/97] Tels: possible longmul overflow in fillbackpack fixed
             Tels: better timing printouts

*/

// set to 0 to get only best backpacks

#define BestOnly 0

// undefine to use other method

#define UseSorting 0

#define EarlyOut 0

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>


 int i;                // for adding items
 int j;                // for timings

 double M;             // backpack size
 unsigned long N;      // number of items

 double MaxM;          // stop if we have reached that M
 int fullstop = 0;

 double Mb;            // backpack size of best solution so far
 unsigned long Nb;     // number of items of best solution so far
 double Vb;            // best backpack value so far

 double Mc;            // current backpack size
 double Mv;            // current backpack value

 double ulSmallestItem;     // for breaking out early

 unsigned long tried;

 int rc;               // return code of fillBackPack;

 struct item {
   double size;             // how big is it?
   double value;            // how much is it worth?
   double quality;          // how much is it worth per size unit? (qual = val / size)
   unsigned long count;     // how much do we have? 0 for unlimited
   unsigned long used;      // how much do we use in backpack? 0 for none, used <= count
   item* next;              // single linked list
 };

// statistics only:
 unsigned long nodes;       // calls to fillbackpack
 unsigned long deepest;     // deepest level of fillpackback()
 unsigned long level;       // current level of fillpackback()

 item* itemlist;            // the items we have to our disposal
 item* finalpack;           // optimal backpack we have found so far

 item* bestfound;           // cache for best next item
// item list stuff

// create sorted list by inserting at proper pos

int addItemSorted (item* head, unsigned long* Ni, double size, double value, unsigned long count)
{
  item* pitem;
  item* pc;

  if (head == NULL)
     {
     return 2; // error, no valid list head
     }
  if (size < ulSmallestItem)
     {
     ulSmallestItem = size;
     }
  pitem = (item*) malloc (sizeof(item));
  pitem->size = size;
  pitem->value = value;
  pitem->count = count;
  pitem->quality = value / size;
  pitem->used = 0;

  pc = head;
  while (pc->next != NULL)
     {
     if ((pc->next->quality < pitem->quality) ||
        ( (pc->next->quality == pitem->quality) && (pc->next->size >= pitem->size) ))
        {
        // but sort by size when quality is equal
        break; // insert before
        }
     pc = pc->next;
     }
  // insert in list just before pc
  // works on empty list and at list end, too
  pitem->next = pc->next; pc->next = pitem;
  // increment count of items in list
  *Ni = *Ni + 1;
  // done
  return 0;
}

// create unsorted list by inserting at list start

int addItemUnsorted (item* head, unsigned long* Ni, double size, double value, unsigned long count)
{
  item* pitem;

  if (head == NULL)
     {
     return 2; // error, no valid list head
     }
  if (size < ulSmallestItem)
     {
     ulSmallestItem = size;
     }
  pitem = (item*) malloc (sizeof(item));
  pitem->size = size;
  pitem->value = value;
  pitem->count = count;
  pitem->quality = value / size;
  pitem->used = 0;

  // insert in list just at start
  // works on empty list and at list end, too
  pitem->next = head->next; head->next = pitem;
  // increment count of items in list
  (*Ni) = (*Ni) + 1;
  // done
  return 0;
}

int addItem (item* head, unsigned long* Ni, double size, double value, unsigned long count)
{
#ifdef UseSorting
   return (addItemSorted (head, Ni, size, value, count));
#else
   return (addItemUnsorted (head, Ni, size, value, count));
#endif
}

// this will only work if flag used AND the current item as used
// to make this work, the all sued counts are increased by one, even
// if the current item is used 0 times, to prevent deadlocks in this search routine
// for the sorted algorithm, you could remove this +1

item* findBestItem (item* head, int iMinSize)
{
  item* pitem;
  item* pc;
  double q,s;

  if (head == NULL)
     {
     return NULL; // error, no valid list head
     }
  pc = head->next; q = 0; s = 0; pitem = NULL;
  while (pc != NULL)
     {
     // find only unsed and fitting items
     if ((pc->used == 0) && (pc->size <= iMinSize))
        {
        if (pc->quality > q)
           {
           q = pc->quality; s = pc->size; pitem = pc;
           }
        else
           {
           if ((pc->quality == q) && ( pc->size < s))
              {
              q = pc->quality; s = pc->size; pitem = pc;
              }
           }
        }
     pc = pc->next;
     }
  // done
  return pitem;
}


int deleteAll (item* head)
{
  item* pc;
  item* pc2;

  // dont del head of list
  pc = head->next;
  while (pc != NULL)
     {
     pc2 = pc->next; // preserve next element
     // free current
     free (pc);
     //next element
     pc = pc2;
     }
  head->next = NULL;
  return 0;
}

int printList (item* head)
{
  item* pc;
  int c;

  if (N > 10)
     {
     printf ("\n Warning: More than 12 items in list, printing only first 12.\n");
     }
  printf ("\n    Size      Value      Quality      Count \n");
  // dont use head of list
  pc = head->next; c = 0;
  while ( (pc != NULL) && (c < 12))
     {
     printf (" %7.4f  %7.4f   %7.4f       %u \n", pc->size, pc->value, pc->quality, pc->count);
     pc = pc->next; c++;
     }
  return 0;
}

int copyAll (item* head1, item* head2)
{
  item* pc;

  deleteAll (head2);
  pc = head1->next;
  while (pc != NULL)
     {
     if (pc->used != 0)
       {
       addItemUnsorted ( head2, &Nb, pc->size, pc->value, pc->count);
       }
     pc = pc->next;
     }
  return 0;
}

int printBackpack ()
{
  item* pc;

  tried++;
  if (Mv > Vb)
     {
     Vb = Mv;
     Mb = M - Mc;
     Nb = 0;
     copyAll (itemlist, finalpack);
     printf ("\n Backpack of size %8.4f (filled: %8.4f, value %8.4f):", M, M-Mc, Mv);
     printf ("\n    Size      Value      Quality     Used \n");
     // dont use head of list
     pc = itemlist->next;
     while (pc != NULL)
        {
        if (pc->used > 1)
           {
           printf (" %9.4f  %9.4f  %9.4f  %i \n", pc->size, pc->value, pc->quality, pc->used - 1);
           }
        pc = pc->next;
        }
     if (Mv >= MaxM)
        {
        fullstop = 1;
        }
     }
  return 1;
}

// end item list stuff

int fillBackpack (item* pItem)
{
  double lTry;
  item* next;

  // some statistics, could be optimized away ;)
  nodes++; level++; if (level > deepest) {deepest = level;}

  if (fullstop) {level--; return 2;} // back out early, and dont do anything
  if (Mc < ulSmallestItem)
     {
     // there wont fit any more items inside
     // backpack must be full
     // even current item wont fit!
     level--;
     return (printBackpack());
     }

#ifdef BestOnly
  // note the <= below !
  if ( Mc * pItem->quality + Mv <= Vb)
     {
     // we cant get a better solution here, so stop
     level--;
     return 2;
    // this only works, if we assume that all items
    // tried after the current have a quality equal or lower
    // than the current item, hence the sorting
     }
#endif

#ifdef UseSorting
  // use only items fitting at least once
  next = pItem;
  do
    {
    next = next->next;
    if (next == NULL)
       {
       break;
       }
    } while (next->size > Mc);
#else
  // fix up for unsorted case
  pItem->used = 1;
  next = findBestItem(itemlist, Mc );
#endif

  // faster variant
  lTry = floor((double)Mc / pItem->size);
  // slower (?)  variant
//  modf ( Mc / pItem->size, &lTry);

  // how many of the current item will fit in?
  // break out condition


  if (next == NULL)
     {
     // end of items reached, backpack cant get fuller
     // fill with maximum count and stop there
     pItem->used = lTry + 1;
     Mc = Mc - (double)lTry * pItem->size;
     Mv = Mv + (double)lTry * pItem->value;
     printBackpack();
     pItem->used = 0;
     Mc = Mc + (double)lTry * pItem->size;
     Mv = Mv - (double)lTry * pItem->value;
     level--;
     return 1;
     }
  for ( ; lTry >= 0; lTry--)
     {
     // insert current item lTry times into backpack
     pItem->used = lTry + 1;
     // and now take next item and free space in backpack that is left
     Mc = Mc - (double)lTry * pItem->size;
     Mv = Mv + (double)lTry * pItem->value;
     rc = fillBackpack (next);
//     fillBackpack (next);
     // reset values
     pItem->used = 0;
     Mc = Mc + (double)lTry * pItem->size;
     Mv = Mv - (double)lTry * pItem->value;
#ifdef EarlyOut
    if (rc == 2)
        {
        break;
        }
#endif
     }
  level--;
  return 0;
}


int main(void)
{
 double q,s; // temp values for creating the items

 time_t timestart, timestop, timesort;
 char timestring[30];
 char* tp;

 time(&timestart);

 printf ("BackPack  v1.1  (c) Copyright by Tels 1997.  All Rights Reserved. \n\n");
 // init lists in memory
 itemlist = new item; itemlist->next = NULL;
 // init list head with dummy values
 itemlist->size = 0; itemlist->value = 0; itemlist->quality = 0;
 itemlist->count = 0;
 finalpack = new item; finalpack->next = NULL;

 ulSmallestItem = 10000000;     // very big

 // add some items

/* M = 37;
 addItem (itemlist, &N, 20, 40, 0);
 addItem (itemlist, &N, 9, 17, 0);
 addItem (itemlist, &N, 4, 7, 0);
*/

 // create items, but distribute the quality and the size good, not size and value !!
/* M = 20000;
 for (i=0; i<1000; i++)
    {
    // dont generate size or value = 0 !!

    // GOOD item generator below
    q = (double)(1000.0*rand()/RAND_MAX + 100) / 100;
    s = (double)(1000.0*rand()/RAND_MAX + 100) / 100;
    addItem (itemlist, &N, s, s*q, 0); if ((i & 0x1F) == 0) {printf ("\r Adding: %i ", i);}

    // BAD item generator below
//    addItem (itemlist, &N, (100*rand()/RAND_MAX + 1), (100*rand()/RAND_MAX + 1), 0); printf ("\r Adding: %i ", i);
    }
*/
 // this creates a backpack with the maximum amount of different
 // items in it...

 M = 0;
 s=3;
 for (i=0; i<20; i++)
    {
    s= s*3;
    }
 q = 2000;
 for (i=0; i<20; i++)
    {
    M = M + s;
    addItem (itemlist, &N, s, s*q, 0); printf ("\r Adding: %i ", i);
    s = (s / 3);
    q -= 1;
    }
  M += 1;

 // german computer magazin c't 11/97 example

/* M = 40;
 addItem (itemlist, &N, 5, 14, 0);
 addItem (itemlist, &N, 3, 8, 0);
 addItem (itemlist, &N, 5, 8, 0);
 addItem (itemlist, &N, 6, 9, 0);
 addItem (itemlist, &N, 7, 10, 0);
 addItem (itemlist, &N, 8, 11, 0);

*/
 // for speed reasons omit this
 printList (itemlist);


 // start with first item
 printf ("Starting...\n");
 time(&timesort);

// for (j=0; j<100; j++)
    {
    Mb = 0; N = 0; Nb = 0; Vb = 0; // currently no solution found
    // start with empty backpack
    Mc = M; Mv = 0; tried = 0; nodes = 0; MaxM = M; deepest = 0; level = 0;
    // use "MaxM = M  - 1" for faster, but not THE best sol;

#ifdef UseSorting
  fillBackpack (itemlist->next);
#else
  fillBackpack (findBestItem (itemlist, M) );
#endif
   deleteAll (finalpack);
   }

 // clean up
 deleteAll (itemlist);
 deleteAll (finalpack);
 printf ("\n Backpacks tried: %u   Calls to fillBackPack(): %u Deepest: %u\n", tried, nodes, deepest);
 printf ("\n Best: (Mstop: %7.2f)", MaxM);

#ifdef UseSorting
  printf (" (Sorting ON,");
#else
  printf (" (Sorting OFF,");
#endif

#ifdef BestOnly
  printf (" Best_only ON,");
#else
  printf (" Best_only OFF,");
#endif

#ifdef EarlyOut
  printf (" Early_out ON)");
#else
  printf (" Early_out OFF)");
#endif

 printf ("\n  Size: %8.2f Filled: %8.2f Value: %8.2f Items: %i\n", M, Mb, Vb, Nb);
 time(&timestop);

 tp = ctime(&timestart); strcpy (timestring,tp); timestring[24] = 0; // kick \n

 printf ("\nDone. (Start: %s, after sort: ", timestring);
 tp = ctime(&timesort); strcpy (timestring,tp); timestring[24] = 0; // kick \n
 printf ("%s\n", timestring);
 tp = ctime(&timestop); strcpy (timestring,tp); timestring[24] = 0; // kick \n
 printf ("       Stop:  %s\n", timestring);
 return 0;
}

