#include <iostream.h>

class Manager; // Forward declaration

class Employee {
public:
  Employee(const char* aName, const Manager* aManager);
  const char* name(void) const;
  const Manager* manager(void) const;
private:
  const char* theName;
  const Manager* theManager;
};

class Secretary; // Forward declaration

class Manager : public Employee {
public:
  Manager(const char* aName,
                 const Manager* aManager,
                 Secretary* aSecretary = 0);
  void setSecretary(Secretary* aSecretary);
  Secretary* secretary(void);
private:
  Secretary* theSecretary;
};

class Secretary : public Employee
{
public:
  Secretary(const char* aName,
            const Manager* aManager);
};

class ProjectLeader : public Employee
{
public:
  ProjectLeader(const char* aName,
                const Manager* aManager);
};

class Engineer : public Employee
{
public:
  Engineer(const char* aname,
           const Manager* aManager,
           const ProjectLeader* aProjectLeader);
  const ProjectLeader* projectLeader(void) const;
private:
  const ProjectLeader* theProjectLeader;
};

Employee::Employee(const char* aName,
                   const Manager* aManager)
  : theName(aName),
    theManager(aManager)
{
}

const char* Employee::name(void) const
{
  return theName;
}

const Manager* Employee::manager(void) const
{
  return theManager;
}

Secretary::Secretary(const char* aName,
                     const Manager* aManager)
  : Employee(aName, aManager) //****
{
}

ProjectLeader::ProjectLeader(const char* aName,
                             const Manager* aManager)
  : Employee(aName, aManager) //****
{
}

Engineer::Engineer(const char* aName,
                   const Manager* aManager,
                   const ProjectLeader* aProjectLeader)
  : Employee(aName, aManager), //****
    theProjectLeader(aProjectLeader)
{
}

const ProjectLeader* Engineer::projectLeader(void) const
{
  return theProjectLeader;
}

Manager::Manager(const char* aName,
                 const Manager* aManager,
                 Secretary* aSecretary)
  : Employee(aName, aManager), //****
    theSecretary(aSecretary)
{
}

void Manager::setSecretary(Secretary* aSecretary)
{
  theSecretary = aSecretary;
}

Secretary* Manager::secretary(void)
{
  return theSecretary;
}

ostream& operator<<(ostream& os, const Employee& e)
{
  if (!os.opfx())
    return os;
  
  cout << '"' << e.name() << "\" is managed by ";
  const Manager* pm = e.manager();
  if (pm) {
    cout << '"' << pm->name() << '"';
  } else {
    cout << "nobody";
  }
  os.osfx();
  
  return os;
}


int main(void)
{
  Manager CEO("Big Boss", 0);
  Secretary sec1st("1st secretary", &CEO);
  CEO.setSecretary(&sec1st);
  
  Secretary shared("shared secretary", &CEO);
  
  Manager middle1("Medium boss 1", &CEO, &shared);
  Manager middle2("Medium boss 2", &CEO, &shared);


  ProjectLeader p("Proj1", &middle1);
  Engineer e("Eng", &CEO, &p); // Managed by CEO, but
                                // work on a project
                                // controlled by p.

  const Manager* pm = CEO.manager();
  cout << "CEO is :" << CEO.name()
       << " whose manager is " 
       << (pm ? pm->name() : "nobody") << endl;
  cout << "The secretary of CEO is "
       <<  CEO.secretary()->name() << endl;
  cout << CEO.secretary()->name() << "'s manager is "
       << CEO.secretary()->manager()->name() << endl;
  
  cout << "The name of middle1 is: " << middle1.name()
       << " whose secretary is "
       << middle1.secretary()->name() << endl;
  cout << "The manager of " 
       << middle1.secretary()->name() << " is "
       << middle1.secretary()->manager()->name() << endl;
  cout << p.name() << " is a project leader managed by "
       << p.manager()->name() << endl;
  cout << e.name() << " is an engineer managed by "
       << e.manager()->name()
       << " and works on a project controlled by "
       << e.projectLeader()->name() << endl;

  cout << endl;
  
  cout << CEO << endl;
  cout << sec1st << endl;
  cout << middle1 << endl;
  cout << middle2 << endl;
  cout << shared << endl;
  cout << p << endl;
  cout << e << endl;

}



